﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ILOG.Concert;
using ILOG.CPLEX;
using System.IO;

namespace KECyclesAndChains
{

   
    public  class RegSet
    {
        public int[] m_array;
        public double m_value;
        public int m_donorIndex = -1;
        public int m_length;
        public int m_type;
        public RegSet()
        {
        }
    }
    public class ExchangeSet
   {
       static public int CYCLE = 1;
       static public int CHAIN = 2;
       public string m_array;

       public ExchangeSet()
       {
       }
       public void AddCycle(int[] array, double value)
       {
           m_array = "c,";
           for (int i = 0; i < array.Length; i++)
               m_array += array[i] + ",";
           m_array += Math.Round(value, 3);
       }
       public void AddChain(int[] array, double value)
       {
           m_array = "n,";
           for (int i = 0; i < array.Length; i++)
               m_array += array[i] + ",";
           m_array += Math.Round(value, 3);
       }
       public void AddDonor(int donorIndex)
       {
           m_array += donorIndex;
       }
       public RegSet GetRegSet()
       {
           RegSet rs = new RegSet();
           int length, i;
           string[] strarr = m_array.Split(',');
           length = strarr.Length;
           int[] arr;
           if (strarr[0] == "c")
           {
               rs.m_type = CYCLE;
               arr = new int[length - 2];
               for (i = 0; i < length - 2; i++)
                   arr[i] = Convert.ToInt32(strarr[i + 1]);
               rs.m_array = arr;
               rs.m_length = length - 2;
               rs.m_value = Convert.ToDouble(strarr[length - 1]);
           }
           else
           {
               rs.m_type = CHAIN;
               arr = new int[length - 2];
               for (i = 0; i < length - 2; i++)
                   arr[i] = Convert.ToInt32(strarr[i + 1]);
               rs.m_array = arr;
               rs.m_length = length - 2;
               rs.m_value = Convert.ToDouble(strarr[length - 1]);
           }
           return rs;
       }
   }



    public class Exchange
    {
        public int m_numOfVariables;
        public int[,] m_vectorAlt = null;
        public bool[] m_altMap = null;
        public List<int> m_altList = null;
        public double m_objVal = 0;
        public int m_xcount = 0;
        public int countFromAlt = 0;
        public int m_type;
        public List<RegSet> m_chosenSets;
        public int m_numOfDistinctChainStarts = 0;
        public int m_numAltDonors = 0;

        public double m_eps = 0.001;
        public double m_delta = 0.001 / 4;
        public List<Node> m_nodeList;
        public int m_maxChainlength;
        public int m_maxCycleLength;


        int[,] m_matrix;
        int m_size;
        public List<ExchangeSet> exchangeSetsList;

        public Exchange(int[,] matrix, int size, int[,] vectorAlt, List<int> altList)
        {
            m_matrix = matrix;
            m_size = size;
            m_nodeList = new List<Node>();
            exchangeSetsList = new List<ExchangeSet>();
            m_vectorAlt = vectorAlt;
            m_numAltDonors = altList.Count();
            m_altMap = new bool[m_size];
            for (int i = 0; i < m_size; i++)
                m_altMap[i] = false;
            for (int i = 0; i < m_numAltDonors; i++)
                m_altMap[altList[i]] = true;
            m_altList = altList;
            //m_size = 30;
        }
        public void Init(int chainLength, int cycleLength)
        {   
            m_maxCycleLength = cycleLength;
            m_maxChainlength = chainLength;
            if (m_chosenSets != null)
                m_chosenSets.Clear();
            if (exchangeSetsList != null)
                exchangeSetsList.Clear();
            //m_size = 8;
        }

        public void WriteToFile(string String, string filename)
        {
            string str = String;
            FileInfo f = new FileInfo("c://test/" + filename + ".txt");
            StreamWriter writer = f.AppendText();
            writer.WriteLine(str);
            writer.Close();
        }



        bool UnrestrictedPopulateByRow(IMPModeler model,
                                      IIntVar[][] var,
                                      IRange[][] rng)
        {
            int count,k;
            m_numOfVariables = m_size*m_size;
            IIntVar[] x = model.IntVarArray(m_numOfVariables, 0, 1);
            var[0] = x;

            double[] objvals = new double[m_numOfVariables];


            for (int i = 0; i < m_size; i++)
                for(int j = 0; j < m_size; j++)
                {
                    if (m_matrix[i, j] == 1)
                    {
                        if (m_altMap[j] == false)
                            objvals[i * m_size + j] = 1;
                    }
                }
            model.AddMaximize(model.ScalProd(x, objvals));
            rng[0] = new IRange[2*m_size];  // number of constraints

            List<int>[] arrayIn = new List<int>[m_size];
            List<int>[] arrayOut = new List<int>[m_size];
            for (int i = 0; i < m_size; i++)
            {
                arrayIn[i] = new List<int>();
                arrayOut[i] = new List<int>();
            }
            for(int i = 0; i < m_size; i++)
                for(int j = 0; j < m_size; j++)
                {
                    if (m_matrix[i, j] == 1)
                    {
                        arrayOut[i].Add(j);
                        arrayIn[j].Add(i);
                    }
                }



            INumExpr[] ieProd;
            INumExpr ieSum;
            for (int i = 0; i < m_size; i++)
            {
                k=0;
                count = arrayIn[i].Count() + arrayOut[i].Count();
                if (count > 0)
                {
                    ieProd = new INumExpr[count];
                    foreach (int j in arrayIn[i])
                    {
                        ieProd[k] = model.Prod(1.0, x[j*m_size+i]);
                        k++;
                    }
                    foreach (int j in arrayOut[i])
                    {
                        ieProd[k] = model.Prod(-1.0, x[i*m_size+j]);
                        k++;
                    }
                    ieSum = model.Sum(ieProd);
                    rng[0][i] = model.AddEq(ieSum, 0);
                }
            }

            for (int i = 0; i < m_size; i++)
            {
                k = 0;
                count = arrayOut[i].Count();
                if (count > 0)
                {
                    ieProd = new INumExpr[count];

                    foreach (int j in arrayOut[i])
                    {
                        ieProd[k] = model.Prod(1.0, x[i * m_size + j]);
                        k++;
                    }
                    ieSum = model.Sum(ieProd);
                    rng[0][i+m_size] = model.AddLe(ieSum, 1);
                }
            }

            return true;
        }

        // prepares the integer program to call  cplex
        bool PopulateByRow(IMPModeler model,
                                      IIntVar[][] var,
                                      IRange[][] rng)
        {
            //  First define the variables, three continuous and one integer
            int i, count, k;
            System.GC.Collect();
            long mem = System.GC.GetTotalMemory(true);
            /*if (mem > 550000000000)
            {
                WriteToFile("To much Memory", "zzz");
                return false;
            }*/

            IIntVar[] x = model.IntVarArray(m_numOfVariables, 0, 1);
            var[0] = x;
            // Objective Function


            double[] objvals = new double[m_numOfVariables];
            List<int>[] arrayRev = new List<int>[m_size];
            for (i = 0; i < m_size; i++)
                arrayRev[i] = new List<int>();
            i = 0; k = 0;
            RegSet set;
            foreach (ExchangeSet setS in exchangeSetsList)
            {
                set = setS.GetRegSet();
                objvals[i] = set.m_value;
                i++;
                for (int j = 0; j < set.m_length; j++)
                    arrayRev[set.m_array[j]].Add(k);
                k++;
            }

            model.AddMaximize(model.ScalProd(x, objvals));
            rng[0] = new IRange[m_size + m_numAltDonors + m_numOfDistinctChainStarts];  // number of constraints


            INumExpr[] ieProd;
            INumExpr ieSum;
            int numOfElements = exchangeSetsList.Count();
            for (i = 0; i < m_size; i++)
            {
                k = 0;
                count = arrayRev[i].Count();
                if (count > 0)
                {
                    ieProd = new INumExpr[count];
                    foreach (int j in arrayRev[i])
                    {
                        ieProd[k] = model.Prod(1.0, x[j]);
                        k++;
                    }
                    ieSum = model.Sum(ieProd);
                    rng[0][i] = model.AddLe(ieSum, 1);
                }
            }
         
            for (i = 0; i < m_size; i++)
                arrayRev[i].Clear();

            // only one chain at once

            if (m_maxChainlength == 0)
                return true;

            int[] numOfChains = new int[m_numAltDonors];
            for (int m = 0; m < m_numAltDonors; m++)
            {
                for (int l = 0; l < m_size; l++)
                {
                    if (m_altMap[l] == false)
                    {
                        if (m_vectorAlt[m, l] == 1)
                        {
                            numOfChains[m]++;
                        }
                    }
                }
            }

            int numOfExchanges = exchangeSetsList.Count();
            for (int m = 0; m < m_numAltDonors; m++)
            {
                INumExpr[] ChainVarVector = new INumExpr[numOfChains[m]];
                INumExpr ChainVarSum;
                i = 0;
                k = 0;
                for (int l = 0; l < m_size; l++)
                {
                    if(m_altMap[l] == false)
                    if (m_vectorAlt[m, l] == 1)
                    {
                        ChainVarVector[k] = model.Prod(1.0, x[numOfExchanges + m * m_size + l]);
                        k++;
                    }
                }
                ChainVarSum = model.Sum(ChainVarVector);
                rng[0][m_size + m] = model.AddLe(ChainVarSum, 1);
            }

            numOfChains = new int[m_size];
            for (i = 0; i < m_size; i++)
                numOfChains[i] = 0;
            foreach (ExchangeSet setS in exchangeSetsList)
            {
                set = setS.GetRegSet();
                if (set.m_type == ExchangeSet.CHAIN)
                    numOfChains[set.m_array[0]]++;
            }

            for (int m = 0; m < m_numAltDonors; m++)
            {
                for (i = 0; i < m_size; i++)
                {
                    if(m_altMap[i] == false)
                    if (m_vectorAlt[m, i] == 1)
                    {
                        numOfChains[i]++;
                    }
                }
            }


            int[] k1 = new int[m_size];
            INumExpr[][] ChainVarVector1 = new INumExpr[m_size][];
            for (i = 0; i < m_size; i++)
            {
                if (numOfChains[i] > 0)
                    ChainVarVector1[i] = new INumExpr[numOfChains[i]];
                k1[i] = 0;
            }

            int k2 = 0;
            foreach (ExchangeSet setS in exchangeSetsList)
            {
                set = setS.GetRegSet();
                if (set.m_type == ExchangeSet.CHAIN)
                {
                    ChainVarVector1[set.m_array[0]][k1[set.m_array[0]]] = model.Prod(1.0, x[k2]);
                    k1[set.m_array[0]]++;
                }
                k2++;
            }


            for (int m = 0; m < m_numAltDonors; m++)
            {
                for (i = 0; i < m_size; i++)
                {
                    if(m_altMap[i] == false)
                    if (m_vectorAlt[m, i] == 1)
                    {
                        ChainVarVector1[i][k1[i]] = model.Prod(-1.0, x[numOfExchanges + m * m_size + i]);
                        k1[i]++;
                    }
                }
           
            }

            k = 0;
            for (i = 0; i < m_size; i++)
            {
                if (numOfChains[i] > 0)
                {
                 
                    INumExpr ChainVarSum = model.Sum(ChainVarVector1[i]);
                    rng[0][m_size + m_numAltDonors + k] = model.AddEq(ChainVarSum, 0);
                    k++;
                }
            }
            if (k != m_numOfDistinctChainStarts)
            {
                int g = 0;
                g++;
            }
            return true;
        }

        public bool RunUnrestricted()
        {
            Cplex cplex = new Cplex();
            System.GC.Collect();
            System.GC.GetTotalMemory(true);
            string  str="";
            try
            {
                IIntVar[][] var = new IIntVar[1][];
                IRange[][] rng = new IRange[1][];
                UnrestrictedPopulateByRow(cplex, var, rng);
                if (cplex.Solve())
                {
                    double[] x = cplex.GetValues(var[0]);
                    m_objVal = cplex.ObjValue;
                    for(int i = 0; i < m_size; i++)
                        for(int j = 0; j < m_size; j++)
                            if (x[i * m_size + j] > 0.7)
                            {
                                str += (i+1) + "\t" + (j+1) + "\t";
                               
                            }


                }
                 WriteToFile(str,"aaa");
            }
            catch (ILOG.Concert.Exception e)
            {
                System.Console.WriteLine("Concert exception caught '" + e + "' caught");
                exchangeSetsList.Clear();
                cplex.End();
                return false;
            }
            return true;
        }

        //finds the chosen allocation (using cplex)
        public bool Run(int bestRes)
        {
            int i, j;
            //return true;
            SetPossibleCycleExchanges();
            SetPossibleChainExchanges();
            m_numOfVariables = exchangeSetsList.Count() + (m_numAltDonors * m_size);

            //Cplex cplex;
            Cplex cplex = new Cplex();
            System.GC.Collect();
            System.GC.GetTotalMemory(true);
            try
            {
                IIntVar[][] var = new IIntVar[1][];
                IRange[][] rng = new IRange[1][];
                RegSet regSet = null;

                // parameters
                //cplex.SetParam(Cplex.IntParam.Probe, 3); // aggressive probing for bin int progs
                //cplex.SetParam(Cplex.IntParam.VarSel, 3); // varsel 3 for strongbranching

                if (PopulateByRow(cplex, var, rng) == false)
                    return false;
                if (cplex.Solve())
                {
                    double[] x = cplex.GetValues(var[0]);
                    //m_xcount = 0;

                    //double[] slack = cplex.GetSlacks(rng[0]);
                    m_objVal = cplex.ObjValue;
                    if (m_objVal > bestRes)
                    {
                        //m_cplexEndTime = cplex.CplexTime;
                        m_chosenSets = new List<RegSet>();

                        for (j = 0; j < exchangeSetsList.Count(); ++j)
                        {
                            if (x[j] > 0.7)
                            {
                                ExchangeSet sss = exchangeSetsList.ElementAt(j);
                                regSet = exchangeSetsList.ElementAt(j).GetRegSet();
                                m_chosenSets.Add(regSet);
                                if (regSet.m_type == ExchangeSet.CHAIN)
                                {
                                    i = 0;
                                    int h = 0;
                                    for (int l = exchangeSetsList.Count() + regSet.m_array[0]; l < x.Length; l += m_size)
                                    {
                                        if (x[l] > 0.7)
                                        {
                                            regSet.m_donorIndex = m_altList[i];
                                            if (m_vectorAlt[i, regSet.m_array[0]] != 1)
                                            {
                                                int g = 0; g++;
                                            }
                                            h++;
                                        }
                                        i++;
                                    }
                                    if (h != 1)
                                    {
                                        int index = exchangeSetsList.Count();
                                        int g = 0; g++;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            catch (ILOG.Concert.Exception e)
            {
               System.Console.WriteLine("Concert exception caught '" + e + "' caught");
               exchangeSetsList.Clear();
               cplex.End();
               return false;
            }
            cplex.ClearModel();
            cplex.EndModel();
            cplex.End();
            return true;
        }


       
        public void WriteSolution(string test)
        {
            string str = test + "\r\n";
            str += "sol = " + m_objVal + "\r\n\r\n";

            bool cc = true;
            if (m_chosenSets != null)
                foreach (RegSet regset in m_chosenSets)
                {
                    str += regset.m_type + "\t";
                    if (regset.m_type == ExchangeSet.CHAIN)
                    {
                        str += (regset.m_donorIndex + 1) + "\t";
                        for (int i = 0; i < regset.m_array.Length-1; i++)
                        {
                            if (m_matrix[regset.m_array[i], regset.m_array[i + 1]] != 1)
                            {
                                cc = false;
                            }
                       
                        }
                        if (m_matrix[regset.m_donorIndex, regset.m_array[0]] != 1)
                        {
                            cc = false;
                        }
                        if (m_matrix[regset.m_array[regset.m_array.Length - 1], regset.m_donorIndex] != 1)
                        {
                            cc = false;
                        }
                    }
                    
                    for (int i = 0; i < regset.m_length; i++)
                    {
                        str += (regset.m_array[i] + 1) + "\t";
                    }
                    str += "\r\n";
                }
            string f = "solution_" + m_maxChainlength + "_" + m_maxCycleLength;
            WriteToFile(str, f);
            
        }

        public void SetPossibleCycleExchanges()
        {
            List<int> cycle = null;
            m_numOfDistinctChainStarts = 0;
            for (int i = 0; i < m_size; i++)
            {
                if (m_altMap[i] == false)
                {
                    cycle = null;
                    cycle = new List<int>();
                    cycle.Add(i);
                    SetCycles(cycle, i);
                    if (cycle != null)
                        cycle.Clear();
                }
            }
        }
        public void SetCycles(List<int> cycle, int current)
        {
            int rc;
            int[] array;
            int length;
            int i, k, start;
            ExchangeSet tmp;
            double value;
            start = cycle[0];
            List<int> tmp_cycle = new List<int>();
            if (cycle.Count() > m_maxCycleLength)
                return;
            start = cycle[0];
            if (cycle.Count() == m_maxCycleLength)
            {
                if (m_matrix[current, start] == 1)
                {
                    length = m_maxCycleLength;
                    array = new int[length];
                    k = 0;
                    foreach (int j in cycle)
                    {
                        array[k] = j;
                        k++;
                    }
                    value = length;
                    tmp = new ExchangeSet();
                    tmp.AddCycle(array, value);
                    exchangeSetsList.Add(tmp);
                }
                return;
            }
            

       
            for (i = 0; i < m_size; i++)
            {
                if (m_altMap[i] == false)
                {
                    foreach (int j in cycle)
                        tmp_cycle.Add(j);
                    rc = 1;
                    if (m_matrix[current, i] == 1 && current != i)
                    {
                        foreach (int j in tmp_cycle)
                        {
                            if (j == i)
                            {
                                if (i == start)
                                    rc = 2;
                                else
                                    rc = 0;
                                break;
                            }
                        }

                        if (rc == 2)
                        {
                            length = tmp_cycle.Count();
                            array = new int[length];
                            k = 0;
                            foreach (int j in tmp_cycle)
                            {
                                array[k] = j;
                                k++;
                            }
                            value = length;
                            tmp = new ExchangeSet();
                            tmp.AddCycle(array, value);
                            exchangeSetsList.Add(tmp);
                        }
                        else if (rc == 1 && i > start)
                        {
                            tmp_cycle.Add(i);
                            SetCycles(tmp_cycle, i);
                        }

                    }
                }
                tmp_cycle.Clear();
            }
        }


        // chains beginning with altruistic donor
        public void SetPossibleChainExchanges()
        {
            int[] array;
            ExchangeSet tmp;
            List<int> chain = null;
            double value;
            m_numOfDistinctChainStarts = 0;
            if (m_maxChainlength == 0)
                return;
            for (int i = 0; i < m_size; i++)
            {
                for (int j = 0; j < m_numAltDonors; j++)
                {
                    chain = null;
                    if (m_altMap[i] == false)
                    {
                        if (m_vectorAlt[j, i] == 1)
                        {
                            m_numOfDistinctChainStarts++;
                            chain = new List<int>();
                            chain.Add(i);
                            array = new int[1];
                            array[0] = i;
                            value = 1;
                            tmp = new ExchangeSet();
                            tmp.AddChain(array, value);
                            exchangeSetsList.Add(tmp);
                            SetChains(chain, i);
                            break;
                        }
                    }
                }
                if (chain != null)
                    chain.Clear();
            }
        }
        public void SetChains(List<int> chain, int current)
        {
            int rc;
            int[] array;
            int length;
            int i, k;
            ExchangeSet tmp;
            double value;
            List<int> tmp_chain = new List<int>();
            if (chain.Count() >= m_maxChainlength)
                return;
            for (i = 0; i < m_size; i++)
            {
                if (m_altMap[i] == false)
                {
                    foreach (int j in chain)
                        tmp_chain.Add(j);
                    rc = 1;
                    if (m_matrix[current, i] == 1 && current != i)
                    {
                        foreach (int j in tmp_chain)
                        {
                            if (j == i)
                            {
                                rc = 0;
                                break;
                            }
                        }
                        if (rc == 1)
                        {
                            tmp_chain.Add(i);
                            length = tmp_chain.Count();
                            array = new int[length];
                            k = 0;
                            foreach (int j in tmp_chain)
                            {
                                array[k] = j;
                                k++;
                            }
                            value = length;
                            tmp = new ExchangeSet();
                            tmp.AddChain(array, value);
                            exchangeSetsList.Add(tmp);
                            SetChains(tmp_chain, i);
                        }

                    }
                    tmp_chain.Clear();
                }
            }
        }


    }
}






/*
 public void SetPossibleCycleExchanges(int maxSize)
        {
            int i, j, k,z count;
            int[] array;
            ExchangeSet tmp;
            double value;

            // two way cycles
            for (i = 0; i < m_size - 1; i++)
            {
                for (j = i + 1; j < m_size; j++)
                {
                    if ((m_matrix[i, j] == 1) && m_matrix[j, i] == 1)
                    {
                        array = new int[2];
                        array[0] = i;
                        array[1] = j;
                        value = 2;
                        tmp = new ExchangeSet();
                        tmp.AddCycle(array, value);
                        exchangeSetsList.Add(tmp);
                    }
                }
            }
            // three way cycles

            if (maxSize >= 3)
            {
                for (i = 0; i < m_size; i++)
                {
                    for (j = i + 1; j < m_size; j++)
                    {
                        for (k = 0; k < m_size; k++)
                        {
                            if ((m_matrix[i, j] == 1) && m_matrix[j, k] == 1 && m_matrix[k, i] == 1)
                            {
                                if (k > i && k != j)  // otherwise this cycle has been found already
                                {
                                    array = new int[3];
                                    array[0] = i;
                                    array[1] = j;
                                    array[2] = k;
                                    value = 3;
                                    tmp = new ExchangeSet();
                                    tmp.AddCycle(array, value);
                                    exchangeSetsList.Add(tmp);
                                }
                            }
                        }
                    }
                }
            }
            if (maxSize == 4)
            {
                for (i = 0; i < m_size; i++)
                {
                    for (j = i + 1; j < m_size; j++)
                    {
                        for (k = i + 1; k < m_size; k++)
                        {
                            for (z = i + 1; z < m_size; z++)
                            {
                                
                            }
                        }
                    }
                }
            }

        }*/

 
