﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows.Forms;

namespace KECyclesAndChains
{

    public class GC
    {
        static Random m_rnd = new Random();
        static public double BLOA = 2;
        static public double BLOB = 1;
        static public double BLOAB = 0;
        static public double BLOO = 3;
        static public int MAX_DONORS = 3;
        static public int MIN_DONORS = 1;
        static public string[] BLOOD_TYPE = { "AB", "B", "A", "O" };
        static public double DONOR_DROP_RATE = 0.1;

        static public string ToBloodType(double Type)
        {
            if (Type == GC.BLOA)
                return "A";
            if (Type == GC.BLOB)
                return "B";
            if (Type == GC.BLOAB)
                return "AB";
            if (Type == GC.BLOO)
                return "O";
            else
                return "none";
        }
        static public double ToBloodType(string Type)
        {
            if (Type == "A")
                return GC.BLOA;
            if (Type == "B")
                return GC.BLOB;
            if (Type == "AB")
                return GC.BLOAB;
            if (Type == "O")
                return GC.BLOO;
            else
                return -1;
        }

        static public int CAUCASIAN = 1; // race indicators
        static public int AFRICANAMERICAN = 2;
        static public int HISPANIC = 3;
        static public int KPD = 0; // PoolType indicators for pairs in Kidney Paired Donation, patients on list, NDDS and deceased donors
        static public int LIST = 1;
        static public int ND = 2;
        static public int DEC = 3; // Demographic assumptions; need real numbers and add specific PRA calc
        static public double[] CAUCASIAN_RANGE = new double[] { 0.972, 0.972, 0.972, 0.972 };
        static public double[] AFRICANAMERICAN_RANGE = new double[] { 0.017 + GC.CAUCASIAN_RANGE[GC.KPD], 0, 017 + GC.CAUCASIAN_RANGE[GC.LIST], 0.017 + GC.CAUCASIAN_RANGE[GC.ND], 0.017 + GC.CAUCASIAN_RANGE[GC.DEC] };
        static public double[] HISPANIC_RANGE = new double[] { 0.011 + GC.AFRICANAMERICAN_RANGE[GC.KPD], 0.011 + GC.AFRICANAMERICAN_RANGE[GC.LIST], 0.011 + GC.AFRICANAMERICAN_RANGE[GC.ND], 0.011 + GC.AFRICANAMERICAN_RANGE[GC.DEC] };
        // blood type and PRA ranges are dependent on demographic
        static public double[] ARANGE = new double[] { 0.34, 0.34, 0.34, 0.34 };
        static public double[] BRANGE = new double[] { 0.14 + GC.ARANGE[GC.KPD], 0.14 + GC.ARANGE[GC.LIST], 0.14 + GC.ARANGE[GC.ND], 0.14 + GC.ARANGE[GC.DEC] };
        static public double[] ORANGE = new double[] { 0.48 + GC.BRANGE[GC.KPD], 0.48 + GC.BRANGE[GC.LIST], 0.48 + GC.BRANGE[GC.ND], 0.48 + GC.BRANGE[GC.DEC] };
        static public double DONORCHANCE = 0.5;
        static public double LOWPRA = 0.95;
        static public double MEDPRA = 0.55;
        static public double HIGHPRA = 0.05;
        static public double[] LOWPRARANGE = new double[] { 0.60, 0.60, 0.60, 0.60 };
        static public double[] MEDPRARANGE = new double[] { 0.25 + GC.LOWPRARANGE[GC.KPD], 0.25 + GC.LOWPRARANGE[GC.LIST], 0.25 + GC.LOWPRARANGE[GC.ND], 0.25 + GC.LOWPRARANGE[GC.DEC] };



        // testing just blood compatability
        static public bool Compatible(double donorBlood, double patientBlood)
        {
            if (donorBlood == GC.BLOA)
            {
                if (patientBlood == GC.BLOA || patientBlood == GC.BLOAB)
                    return true;
                return false;
            }
            else if (donorBlood == GC.BLOB)
            {
                if (patientBlood == GC.BLOB || patientBlood == GC.BLOAB)
                    return true;
                return false;
            }
            else if (donorBlood == GC.BLOAB)
            {
                if (patientBlood == GC.BLOAB)
                    return true;
                return false;
            }
            else if (donorBlood == GC.BLOO)
                return true;
            return false;
        }
        public static bool Compatible(Node nodePatient, Donor donor, bool CheckPRA)
        {
            if (GC.Compatible(donor.m_bloodType, nodePatient.m_bloodTypePatient) == true)
            {
                //if (GC.CHECK_PRA_AFTER == false || GC.VIRTUAL_XMATCH)
                //  CheckPRA = true;

                if (CheckPRA)
                {
                    if (nodePatient.CrossMatch() == true)
                        return true;
                }
                else return true;
            }
            return false;
        }
        public static bool SpecialCompatible(Node nodePatient, Donor donor)
        {
            if (GC.SpecialCompatible(donor.m_bloodType, nodePatient.m_bloodTypePatient) == true)
            {
                if (nodePatient.CrossMatch() == true)
                    return true;
            }
            return false;
        }

        // check compatibility according the real rules today
        public static bool SpecialCompatible(double donorBlood, double patientBlood)
        {
            if (donorBlood == GC.BLOA)
            {
                if (patientBlood == GC.BLOA || patientBlood == GC.BLOAB)
                    return true;
                return false;
            }
            else if (donorBlood == GC.BLOB)
            {
                if (patientBlood == GC.BLOB)
                    return true;
                return false;
            }
            else if (donorBlood == GC.BLOAB)
            {
                if (patientBlood == GC.BLOAB)
                    return true;
                return false;
            }
            else if (donorBlood == GC.BLOO && patientBlood == GC.BLOO)
                return true;
            return false;
        }

      
        static public double NewPRA(int Demographic)
        {
            double tmp = m_rnd.NextDouble();
            if (tmp < GC.LOWPRARANGE[Demographic])
                return GC.LOWPRA;
            if (tmp < GC.MEDPRARANGE[Demographic])
                return GC.MEDPRA;
            return GC.HIGHPRA;
        }

        // randomize a new integer
        static public int GetInt(double num)
        {
            double tmp = m_rnd.NextDouble();
            tmp = tmp * num;
            int tmp1 = (int)tmp + 1;
            return tmp1;
        }
        static public int GetInt(int a, int b)
        {
            double tmp = m_rnd.NextDouble();
            tmp = 0.5 + a + (b - a) * tmp;
            int tmp1 = (int)tmp;
            return tmp1;
        }
        static public double GetDouble()
        {
            return m_rnd.NextDouble();
        }
        static public int GetDemographic(int PoolType)
        {
            double tmp = m_rnd.NextDouble();
            if (tmp < GC.CAUCASIAN_RANGE[PoolType])
                return GC.CAUCASIAN;
            if (tmp < GC.AFRICANAMERICAN_RANGE[PoolType])
                return GC.AFRICANAMERICAN;
            return GC.HISPANIC;
        }

    }

    public class PeopleManager
    {
        public PeopleManager()
        {
        }

        public Node CreatePatientWithIncDonors()
        {
            Node node = null;
            Donor donor;
            int tmp1;
            for (int i = 0; i < 1; i++)
            {
                node = new Node();
            newnode:
                node.Initialize();
                tmp1 = GC.GetInt(GC.MIN_DONORS, GC.MAX_DONORS);
                for (int j = 0; j < tmp1; j++)
                {
                    donor = node.GenerateDonor();
                    if (GC.Compatible(node, donor, true) == true)
                        goto newnode;
                }

                if (node.m_numOfDonors > 1 && GC.DONOR_DROP_RATE != 0)
                { // drop so average is 1.2
                    double tmp2 = GC.GetDouble();
                    if (tmp2 < GC.DONOR_DROP_RATE)
                    {
                        if (node.m_numOfDonors > 2)
                        {
                            for (int k = 2; k < node.m_numOfDonors; k++)
                            {
                                node.m_donor[k] = null;
                            }
                            node.m_numOfDonors = 2;
                        }
                    }
                    else
                    {
                        for (int k = 1; k < node.m_numOfDonors; k++)
                            node.m_donor[k] = null;
                        node.m_numOfDonors = 1;
                    }
                }

            }
            return node;
        }

    }

    public class Donor
    {
        public int m_index;
        public double m_bloodType;
        public Donor()
        {
        }
        public static double GetNewBloodType(int Demographic)
        {
            double tmp = GC.GetDouble();
            //change so it will be according to patient
            if (tmp < GC.ARANGE[Demographic])
                return GC.BLOA;
            else if (tmp < GC.BRANGE[Demographic])
                return GC.BLOB;
            else if (tmp < GC.ORANGE[Demographic])
                return GC.BLOO;///////CHANGE BLOOD TYPE
            else
                return GC.BLOAB;
        }
    }

    public class Node
    {
        public int m_index;
        public double m_PRA;
        public double m_bloodTypePatient;
        public List<double> m_donorsType = null;   // for concise use, just hold the blood types
        public List<Donor> m_donor = null; // for more extensive simulations to have donors attributes
        public int m_numOfDonors;
        public int m_demographic;

        public Node(int index, double bloodType, double PRA)
        {
            m_donorsType = new List<double>(); // use just to hold indices of donors
            m_donor = new List<Donor>();  // use only for simulations...
            m_index = index;
            m_bloodTypePatient = bloodType;
            m_PRA = PRA;
            m_numOfDonors = 0;
            m_demographic = 0;
        }

        public Node()
        {
            m_donorsType = new List<double>(); // use just to hold indices of donors
            m_donor = new List<Donor>();  // use only for simulations...
            m_numOfDonors = 0;
            m_demographic = 0;
        }

        public void Initialize()
        {
            m_demographic = GC.GetDemographic(GC.LIST);
            m_PRA = GC.NewPRA(m_demographic);
            SetBloodType(m_demographic);
        }

        public void SetBloodType(int Demographic)
        {
            double tmp = GC.GetDouble();
            if (tmp < GC.ARANGE[Demographic])
            {
                m_bloodTypePatient = GC.BLOA;
                return;
            }
            if (tmp < GC.BRANGE[Demographic])
            {
                m_bloodTypePatient = GC.BLOB;
                return;
            }
            if (tmp < GC.ORANGE[Demographic])
            {
                m_bloodTypePatient = GC.BLOO;///////CHANGE BLOOD TYPE
                return;
            }
            m_bloodTypePatient = GC.BLOAB;
        }

        public Donor GenerateDonor()
        {
            if (m_numOfDonors == GC.MAX_DONORS)
            {
                throw new System.ArgumentException("to many donors for one node", "original");
            }
            m_numOfDonors += 1;
            m_donor[m_numOfDonors - 1] = new Donor();
            m_demographic = GC.GetDemographic(GC.KPD); // paired demographic
            double donorBloodType = Donor.GetNewBloodType(m_demographic);
            m_donor[m_numOfDonors - 1].m_bloodType = donorBloodType;
            return m_donor[m_numOfDonors - 1];
        }
        public bool CrossMatch()
        {
            double tmp = GC.GetDouble();
            if (tmp > this.m_PRA)
                return false;
            return true;
        }
    }


    
     
    public class Params
    {
        public int[,] m_matrix = null;
        public int[,] m_altVector = null;
        public List<Node> m_NodeList = null;
        public List<int> m_AltList = null;
        public int m_size;
        int m_numOfAlts;
 
        public Params()
        {
            m_NodeList = new List<Node>();
            m_AltList = new List<int>();
        }

        public Node FindNodeWithIndex(int Index)
        {
            foreach (Node node in m_NodeList)
            {
                if (node.m_index == Index)
                    return node;
            }
            throw new System.ArgumentException("node not found", "original");
        }


        public void  ReadCompatMatrix(string CompatMatrixFilename)
        {
            List<string> list = new List<string>();
            using (StreamReader reader = new StreamReader("c://RealData/" + CompatMatrixFilename + ".txt"))
            {
                string line;
                while ((line = reader.ReadLine()) != null)
                {
                    list.Add(line);          // Add to list.
                }
            }

            for (int i = list.Count() - 1; i > -1; i--) // clean up the read
            {
                string String = list.ElementAt(i);
                if (String.Contains("1") == false && String.Contains("0") == false)
                    list.Remove(String);
            }

            m_size = list.Count();
            m_matrix = new int[m_size, m_size];

            for (int i = 0; i < m_size; i++)
            {
                string str = list.ElementAt(i);
                string[] s = str.Split('\t');
                for (int j = 0; j < m_size; j++)
                    m_matrix[j, i] = Convert.ToInt32(s[j]);
            }
        }

        public void ReadListOfNodes(string ListNodesFileName)
        {

            // note double listing of altruistic nodes; need to remove altruistic lines from compat matrix.
            List<Node>[] tmplist = new List<Node>[3];
            for (int i = 0; i < 3; i++)
                tmplist[i] = new List<Node>();
            List<string> list = new List<string>();
            List<int[]> ListOfDoubles = new List<int[]>();
            using (StreamReader reader = new StreamReader("c://RealData/" + ListNodesFileName + ".txt"))
            {
                string line;
                while ((line = reader.ReadLine()) != null)
                {
                    if (line.Length > 0)
                        list.Add(line);          // Add to list.
                }
            }

            bool listingnodes = true;
            bool doubles = true;

            for (int i = 0; i < list.Count(); i++)
            {
            start:
                string str = list.ElementAt(i);
                if (str == "DOUBLES")
                {
                    listingnodes = false;
                    i++;
                    goto start;
                }
                if (str == "ALTRUISTIC")
                {
                    doubles = false;
                    i++;
                    goto start;
                }

                if (listingnodes)
                {
                    int param = 0;
                    string[] s = str.Split('\t');
                    int index = Convert.ToInt32(s[param]) - 1;
                    double pra = -1;
                    double bloodtype = 0;
                    param++;
                    if (s[param] != "")
                        pra = (Convert.ToDouble(s[param]) / 100);
                    param++;
                    if (s[param] != "")
                        bloodtype = GC.ToBloodType(s[param]);
                    param++;
                    Node node = new Node(index, bloodtype, pra);
                    double donortype = GC.ToBloodType(s[param].Split()[0]);
                    node.m_donorsType.Add(donortype);
                    m_NodeList.Add(node);
                }

                if (listingnodes == false && doubles)
                {
                    string[] s = str.Split('\t');
                    int index1 = Convert.ToInt32(s[0]) - 1;
                    int index2 = Convert.ToInt32(s[1]) - 1;

                    int[] indices = { index1, index2 };
                    ListOfDoubles.Add(indices);
                }
                if (listingnodes == false && doubles == false)
                {
                    int index = Convert.ToInt32(str) - 1;
                    Node node = FindNodeWithIndex(index);
                    m_AltList.Add(index);
                    m_NodeList.Remove(node);
                }
            }
            // go through and take out the (non alt) doubles
            foreach (int[] indices in ListOfDoubles)
            {
                int index1 = indices[0];
                int index2 = indices[1];
                bool altdouble = false;

                foreach (int donor in m_AltList)
                {
                    if (donor == index1 || donor == index2)
                    {
                        altdouble = true;
                        break;
                    }
                }

                if (altdouble == false)
                {
                    Node node1 = FindNodeWithIndex(index1);
                    Node node2 = FindNodeWithIndex(index2);
                    node1.m_donorsType.Add(node2.m_donorsType[0]);
                    m_NodeList.Remove(node2);
                }
            }
            m_numOfAlts = m_AltList.Count();

            m_altVector = new int[m_numOfAlts, m_size+1];
            int xx = 0;
            for (int j = 0; j < m_numOfAlts; j++)
            {
                for (int i = 0; i < m_size; i++)
                {
                    if (m_AltList.Contains(i) == false)
                    {
                        m_altVector[j, i] = m_matrix[m_AltList[j], i];
                        if (m_matrix[i, m_AltList[j]] != 1)
                        {
                            xx++;
                            m_matrix[i, m_AltList[j]] = 1;
                        }
                    }
                   
                }
            }
            xx = xx;
            
        }
    }
}
